﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using UnityEngine;

public class SplineTraverser : MonoBehaviour, IAreaEntity
{
    private const int PointBuffer = 4;

    [SerializeField] private RageSpline rageSpline;
    [SerializeField] private float pointDistance = 0.1f;
    [SerializeField] private float speed = 0.1f;
    [SerializeField] private int density = 9;
    [SerializeField] private float tightness = 0.25f;
    [SerializeField] private int addPointCount = 4;
    [SerializeField] private int extraFollowerPoints = 0;

    [SerializeField] private float randomPointMinDistance = 6;
    [SerializeField] private float randomPointMaxDistance = 9;
    [SerializeField] private float randomPointMaxAngleDelta = 90f;

    private RageCurve spline;
    private SplinePath splinePath;

    private float currentStartPosition;
    private int followerCount;
    private float?[] followerSplineIndizes;

    public Rect FullAssignedArea { get; set; }

    private void Awake()
    {
        if (rageSpline == null)
        {
            spline = new RageCurve(new Vector3[0], new Vector3[0], new bool[0], new float[0]);
        }
        else
        {
            spline = rageSpline.spline;
        }

        spline.points = new RageSplinePoint[0];

        splinePath = new SplinePath(spline, density, PointBuffer, tightness);

        followerSplineIndizes = new float?[extraFollowerPoints];

        currentStartPosition = 0f;
    }

    public void Initialize(Rect area)
    {
        FullAssignedArea = area;
        RemoveAndAddPoints();
    }

    private void Start()
    {
        RemoveAndAddPoints();
    }

    public void RandomizePoints()
    {
        spline.points = new RageSplinePoint[0];
        RemoveAndAddPoints();
    }

    private void FixedUpdate()
    {
        currentStartPosition = splinePath.MoveSplineIndex(currentStartPosition, speed * Time.deltaTime, 1);
        ResetFollowerCache();

        /*
        if (currentStartPosition >= splinePath.PointCount - 1)
            currentStartPosition = 0f;

        if (currentStartPosition < 0f)
            currentStartPosition = splinePath.PointCount - 1 - 0.01f;
         */
    }

    private void Update()
    {
    }

    private void ResetFollowerCache()
    {
        for (int i = 0; i < followerSplineIndizes.Length; i++)
            followerSplineIndizes[i] = null;
    }

    public Vector2 GetPosition(int followerIndex)
    {
        if (splinePath.PointCount == 0)
        {
            RemoveAndAddPoints();
        }

        var splineIndex = GetSplineIndex(followerIndex);
        return splinePath.GetSplinePositionBySplineIndex(splineIndex);
    }

    private float GetSplineIndex(int followerIndex)
    {
        if (followerIndex <= 0)
            return currentStartPosition;

        if (followerSplineIndizes[followerIndex] != null)
            return followerSplineIndizes[followerIndex].Value;

        var result = splinePath.MoveSplineIndex(GetSplineIndex(followerIndex - 1), pointDistance, 1);
        if (result == (splinePath.PointCount - 1))
        {
            RemoveAndAddPoints();
            result = splinePath.MoveSplineIndex(GetSplineIndex(followerIndex - 1), pointDistance, 1);
        }

        followerSplineIndizes[followerIndex] = result;
        return result;
    }

    public Vector2[] GetUncachedPositions(int count)
    {
        var positions = new Vector2[count];

        var index = currentStartPosition;

        positions[0] = splinePath.GetSplinePositionBySplineIndex(index);

        for (int i = 1; i < count; i++)
        {
            var shiftedIndex = splinePath.MoveSplineIndex(index, pointDistance, 1);
            if (shiftedIndex == (splinePath.PointCount - 1))
            {
                RemoveAndAddPoints();
                shiftedIndex = splinePath.MoveSplineIndex(index, pointDistance, 1);
            }

            index = shiftedIndex;
            positions[i] = splinePath.GetSplinePositionBySplineIndex(index);
        }

        return positions;
    }

    private void RemoveAndAddPoints()
    {
#if UNITY_EDITOR
        if (FullAssignedArea.width == 0)
        {
            FullAssignedArea = ScreenManager.Instance.Rect;
        }
#endif

        var newCurrentStartPosition = splinePath.RemoveUnusedPoints(currentStartPosition - density * PointBuffer) + density * PointBuffer;
        if (currentStartPosition != newCurrentStartPosition)
        {
            currentStartPosition = newCurrentStartPosition;
            ResetFollowerCache();
        }

        do
        {
            for (int i = 0; i < addPointCount; i++)
            {
                splinePath.AddRandomPoint(randomPointMinDistance, randomPointMaxDistance, randomPointMaxAngleDelta, FullAssignedArea);
                //spline.AddRageSplinePoint(rageSpline.spline.points.Length, rect.RandomPosition(10), Vector3.zero, Vector3.zero, 1f, true);
            }
        } while (spline.points.Length < PointBuffer);

        splinePath.RefreshControlPoints();
        splinePath.RefreshPoints();

        if (rageSpline != null)
        {
            rageSpline.VertexDensity = density;
            rageSpline.RefreshMesh(true, true, false);
        }
    }

    public int AddFollower()
    {
        /*
        if (followerCount > 0)
            currentStartPosition = MoveSplineIndex(currentStartPosition, pointDistance, 1);
         */

        followerCount++;

        followerSplineIndizes = new float?[followerCount + extraFollowerPoints];

        return followerCount - 1;
    }

    public void RemoveFollower()
    {
        followerCount--;

        if (followerCount == 0)
        {
            Destroy(gameObject);
        }
    }
}
