﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using UnityEngine;

public class ShapedIngameEntity : MonoBehaviour, IIngameEntity, IAreaEntity
{
    [SerializeField] private bool solid = true;
    [SerializeField] private AbstractShape mainColorShape;
    [SerializeField] private bool ignoreBorder;
    [SerializeField] private float borderOffset;

    private CollisionManager collisionManager;

    private AbstractShape[] shapes;

    public event Action EventCollidedWithSolid;

    private List<IngameEntityPoint> catchablePoints;
    private bool _ghostlike;
    private bool wasAlreadyStarted;

    private Rect _assignedArea;
    private Rect _useableArea;

    private void Awake()
    {
        collisionManager = CollisionManager.Instance;

        shapes = GetComponentsInChildren<AbstractShape>().Where(shape => !shape.IgnoreShape).ToArray();
        catchablePoints = new List<IngameEntityPoint>();

        if (mainColorShape == null)
            mainColorShape = GetComponentInChildren<AbstractShape>();
    }

    public void Initialize(Rect fullArea, float areaBorder, int groupId)
    {
        AreaBorder = areaBorder;
        FullAssignedArea = fullArea;
        GroupId = groupId;
    }

    public int GroupId { get; private set; }
    public float AreaBorder { get; private set; }

    private void Start()
    {
        if (solid)
            RegisterEntity();

        wasAlreadyStarted = true;

#if UNITY_EDITOR
        if (FullAssignedArea.width == 0)
        {
            FullAssignedArea = ScreenManager.Instance.Rect;
        }
#endif
    }

    private void OnEnable()
    {
        if (wasAlreadyStarted)
        {
            if (solid)
                RegisterEntity();
        }

        //screenManager.EventAspectRatioChanged += OnAspectRatioChanged;
    }

    private void OnDisable()
    {
        if (solid)
            UnregisterEntity();

        //screenManager.EventAspectRatioChanged -= OnAspectRatioChanged;
    }

    public bool MoveTowards(Vector2 target, float maxDelta)
    {
        var delta = target - (Vector2) transform.position;
        if (delta.sqrMagnitude > maxDelta * maxDelta)
        {
            delta = Vector2.ClampMagnitude(delta, maxDelta);
        }

        return Move(delta);
    }

    public bool MoveTo(Vector2 target)
    {
        return Move(target - (Vector2) transform.position);
    }

    public bool Move(Vector2 delta)
    {
        var newPosition = transform.position + (Vector3) delta;

        var clampedPosition = ClampInUseableArea(newPosition);

        transform.position = clampedPosition;

        var fullMove = newPosition == clampedPosition;

        if (!fullMove && (EventCollidedWithSolid != null))
            EventCollidedWithSolid();

        return fullMove;
    }

    public Vector3 ClampInUseableArea(Vector3 position)
    {
        return new Vector3(Mathf.Clamp(position.x, UseableArea.xMin, UseableArea.xMax),
                                          Mathf.Clamp(position.y, UseableArea.yMin, UseableArea.yMax),
                                          transform.position.z);
    }

    public bool TestMove(Vector2 delta)
    {
        return UseableArea.Contains((Vector2) transform.position + delta);
    }

    public void TeleportTo(Vector2 target)
    {
        transform.position = new Vector3(target.x, target.y, transform.position.z);
    }

    public void RotateDegCCW(float angleDeg)
    {
        transform.Rotate(0, 0, angleDeg);
    }

    public void RotateTowardsDeg(float targetAngleDeg, float maxDegDelta)
    {
        Quaternion targetRotation = Quaternion.Euler(0, 0, targetAngleDeg);
        transform.rotation = Quaternion.RotateTowards(transform.rotation, targetRotation, maxDegDelta);
    }

    public bool HasReachedAngleDeg(float targetAngleDeg)
    {
        return MathUtil.IsSameAngle(targetAngleDeg, RotationDeg);
    }

    public float RotationDeg
    {
        get { return transform.rotation.eulerAngles.z; }
        set { transform.rotation = Quaternion.Euler(0, 0, value); }
    }

    public Vector2 GetPointWorldPosition(int shapeIndex, int pointIndex)
    {
        return shapes[shapeIndex].GetPointWorldPosition(pointIndex);
    }

    /*
    private void OnAspectRatioChanged()
    {
        var screenRect = screenManager.Rect;

        var clampedPosition = new Vector3(Mathf.Clamp(transform.position.x, screenRect.xMin, screenRect.xMax),
                                          Mathf.Clamp(transform.position.y, screenRect.yMin, screenRect.yMax),
                                          transform.position.z);

        transform.position = clampedPosition;
    }
     */

    /*
    public List<Vector2> CreateCatchablePointsWorldPositionsList()
    {
        /*
        var points = new List<Vector2>();
        foreach (var shape in shapes)
        {
            var pointCount = shape.PointCount;
            for (int pointIndex = 0; pointIndex < pointCount; pointIndex++)
            {
                if (shape.IsPointCatchable(pointIndex))
                    points.Add(shape.GetPointWorldPosition(pointIndex));
            }
        }
        return points;
         */
    /*
        return catchablePoints.Select(point => point.WorldPosition).ToList();
    }
     */

    public IEnumerable<IngameEntityPoint> CatchablePoints { get { return catchablePoints; } }

    /*
    public Vector2 GetCatchableCenterWorldPosition()
    {
        var center = new Vector2();
        int count = 0;

        foreach (var point in catchablePoints)
        {
            center += point.WorldPosition;
            count++;
        }
        /*
        foreach (var shape in shapes)
        {
            var pointCount = shape.PointCount;
            for (int pointIndex = 0; pointIndex < pointCount; pointIndex++)
            {
                if (shape.IsPointCatchable(pointIndex))
                {
                    center += (Vector2) shape.GetPointWorldPosition(pointIndex);
                    count++;
                }
            }
        }
         */
    /*
        center /= count;

        return center;
    }
    */

    public bool CheckCollision(Vector2 previousWorldPosition, Vector2 currentWorldPosition)
    {
        foreach (var shape in shapes)
        {
            if (shape.CheckCollision(previousWorldPosition, currentWorldPosition))
                return true;
        }

        return false;
    }

    private void RegisterEntity()
    {
        collisionManager.RegisterEntity(this);

        for (int shapeIndex = 0; shapeIndex < shapes.Length; shapeIndex++)
        {
            var shape = shapes[shapeIndex];
            var pointCount = shape.PointCount;
            for (int pointIndex = 0; pointIndex < pointCount; pointIndex++)
            {
                var ingameEntityPoint = new IngameEntityPoint(this, shapeIndex, pointIndex);
                catchablePoints.Add(ingameEntityPoint);
                collisionManager.RegisterPoint(ingameEntityPoint);
            }
        }
    }

    private void UnregisterEntity()
    {
        collisionManager.UnregisterEntityAndPoints(this);
        catchablePoints.Clear();
    }

    public float BoundingRadius { get { return shapes.Select(shape => shape.BoundingRadius).Max(); } }

    public Rect FullAssignedArea
    {
        get { return _assignedArea; }
        set { if (_assignedArea == value)
                return;
            
            _assignedArea = value;

            UpdateUseableArea();
        }
    }

    public float BorderOffset
    {
        get { return borderOffset; }
        set
        {
            if (borderOffset == value)
                return;

            borderOffset = value;

            UpdateUseableArea();
        }
    }

    private void UpdateUseableArea()
    {
        var offset = borderOffset;
        if (!ignoreBorder)
        {
            offset += AreaBorder;
        }

        UseableArea = _assignedArea.CreateExtendedRect(-BoundingRadius - offset);
    }

    public Rect UseableArea
    {
        get { return _useableArea; }
        private set { if (_useableArea == value)
                return;
            
            _useableArea = value;

            transform.position = ClampInUseableArea(transform.position);
        }
    }

    public Color MainColor { get { return mainColorShape.DefaultOutlineColor; } }
}
