﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using UnityEngine;

public class GuiManager : SingletonMonoBehaviour<GuiManager>
{
    [SerializeField] private float fontSizeHudText = 0.045f;
    [SerializeField] private float fontSizeIngameInfo = 0.034f;
    [SerializeField] private float fontSizeLevelTitle = 0.068f;
    [SerializeField] private float fontSizeLevelCountdown = 0.034f;

    [SerializeField] private Color mainGuiColor = Color.white;
    [SerializeField] private Color positiveInfo = Color.white;
    [SerializeField] private Color negativeInfoColor = Color.red;

    [SerializeField] private Vector2 scoreOffset;
    [SerializeField] private Color pointsEntityColor = new Color(1f, 1f, 1f, 0.5f);
    [SerializeField] private float pointsTimeTillVanish = 1f;
    [SerializeField] private float infoTimeTillVanish = 1f;
    [SerializeField] private float levelClearTimeTillVanish = 2.5f;
    [SerializeField] private float defaultVanishTime = 0.5f;

    [SerializeField] private float comboBrokenTextOffsetY = 0.05f;
    [SerializeField] private float levelClearTextOffsetY = 0.05f;
    [SerializeField] private float levelPlacementTextOffsetY = 0.1f;

    [SerializeField] private float levelStartingNumberOffset = 0.01f;
    [SerializeField] private float levelStartingNumberDistance = 0.01f;
    [SerializeField] private float timeDistanceBetweenCountdown = 0.5f;
    [SerializeField] private float levelStartingTimeOffset;

    private VisualSettings visualSettings;

    private Camera mainCamera;

    private List<GuiVanishingText> texts;

    private Vector2 totalScoreGUIPosition;
    private Vector2 lastLevelEndedScorePosition;

    private List<GuiButton> mouseOverButtons;
    private GuiButton _currentlyHoveredButton;

    public Font Font { get { return visualSettings.FontDefault; } }

    public bool Stopped { get; set; }

    public float ScoreLevelToTotalTransfer { get; set; }

    public bool ShowExtraLifes { get; set; }
    public float? PlayTimeLeft { get; set; }

    public float FontSizeIngameInfo { get { return fontSizeIngameInfo; } }
    public Color NegativeInfoColor { get { return negativeInfoColor; } }

    private void Awake()
    {
        visualSettings = VisualSettings.Instance;

        mainCamera = Camera.main;

        texts = new List<GuiVanishingText>();

        mouseOverButtons = new List<GuiButton>();
    }

    private void Update()
    {
        if (mouseOverButtons.Count > 1)
        {
            UpdateHoveredMouseButtonAmongstMultiple();
        }

        if (Stopped)
            return;

        foreach (var text in texts)
        {
            text.Update();
        }

        texts.RemoveAll(text => text.Vanished);
    }

    public void OnGUI()
    {
        if (Stopped)
            return;

        if (PlayTimeLeft == 0)
        {
            UnityHelper.SetFont(Font, fontSizeLevelTitle);
            ShowTextLocalized("IngameGui.PlayTimeIsUp", new Vector2(Screen.width / 2f, Screen.height / 2f), UnityHelper.AnchorCenter, mainGuiColor);
        }

        foreach (var text in texts)
        {
            text.Draw();
        }
    }

    public void ShowLevelTitle(string title, float startingTime)
    {
        startingTime += levelStartingTimeOffset;

        var center = new Vector2(Screen.width / 2f, Screen.height / 2f);

        texts.Add(new GuiVanishingText(title, center, UnityHelper.AnchorCenter, positiveInfo, Font, fontSizeLevelTitle, startingTime - defaultVanishTime, defaultVanishTime));

        for (int i = 0; i < 3; i++)
        {
            texts.Add(new GuiVanishingText((i+1).ToString(), new Vector2(center.x, center.y + (levelStartingNumberOffset + levelStartingNumberDistance * i) * Screen.height), UnityHelper.AnchorCenter,
                positiveInfo, Font, fontSizeLevelCountdown, (startingTime - (i + 1) * timeDistanceBetweenCountdown) - defaultVanishTime, defaultVanishTime));
        }
    }

    /*
    public void ShowEndingMessage()
    {
        var center = new Vector2(Screen.width / 2f, Screen.height / 2f);
        texts.Add(new GuiVanishingText("Thanks for playing! :)", center, UnityHelper.AnchorCenter, levelStartingColor, Font, fontSizeLevelTitle, float.PositiveInfinity, defaultVanishTime));
        texts.Add(new GuiVanishingText("[Press Space to restart the game]", center + new Vector2(0, levelStartingNumberOffset), UnityHelper.AnchorCenter, levelStartingColor, Font, fontSizeLevelCountdown, float.PositiveInfinity, defaultVanishTime));
    }
     */

    public void ShowEntityPoints(Vector2 worldPosition, int score)
    {
        var screenPosition = GetScreenPosition(worldPosition);
        var text = Get("IngameGui.EntityPoints", score);

        texts.Add(new GuiVanishingText(text, screenPosition, UnityHelper.AnchorCenter, pointsEntityColor, Font, fontSizeIngameInfo, pointsTimeTillVanish, defaultVanishTime));
    }

    public void ShowTotalPoints(int totalScore, float bonusMultiplier, float homogeneousGroupMultiplier, bool combo, Color? color)
    {
        var text = Get("IngameGui.TotalPoints", totalScore);
        if (bonusMultiplier > 0)
        {
            var bonusRounded = 1 + Mathf.Round(bonusMultiplier * 10) / 10f;
            text += String.Format(" x{0}", bonusRounded.ToString("0.0"));
        }

        if (homogeneousGroupMultiplier > 1f)
        {
            text = String.Format("({0}) x{1}", text, homogeneousGroupMultiplier);
        }

        if (combo)
        {
            text = Get("IngameGui.TotalPointsCombo") + "\n" + text;
        }

        texts.Add(new GuiVanishingText(text, totalScoreGUIPosition, UnityHelper.AnchorCenter, color ?? positiveInfo, Font, fontSizeIngameInfo, pointsTimeTillVanish, defaultVanishTime, TextAlignment.Center));
    }

    public void ShowLevelDone()
    {
        var cheersString = Get("IngameGui.Cheers");
        var cheersArray = cheersString.Split(',');
        var cheer = cheersArray.RandomElement().Trim();

        var text = Get("IngameGui.LevelClear", cheer);
        texts.Add(new GuiVanishingText(text, totalScoreGUIPosition + new Vector2(0, levelClearTextOffsetY * Screen.height), UnityHelper.AnchorCenter, positiveInfo, Font, fontSizeIngameInfo, levelClearTimeTillVanish, defaultVanishTime, TextAlignment.Center));

        lastLevelEndedScorePosition = totalScoreGUIPosition;
    }

    public void ShowLevelPlacement(int total)
    {
        var text = Get("IngameGui.LevelPlacement", total);
        texts.Add(new GuiVanishingText(text, lastLevelEndedScorePosition + new Vector2(0, levelPlacementTextOffsetY * Screen.height), UnityHelper.AnchorCenter, positiveInfo, Font, fontSizeIngameInfo, levelClearTimeTillVanish, defaultVanishTime, TextAlignment.Center));
    }

    public void ShowNetDestroyed(Vector2 worldPosition)
    {
        var screenPosition = GetScreenPosition(worldPosition);
        AddGuiVanishingTextLocalized("IngameGui.NetDestroyed", screenPosition, UnityHelper.AnchorUpperLeft, negativeInfoColor, Font, fontSizeIngameInfo, infoTimeTillVanish, defaultVanishTime);
    }

    public void ShowComboBroken(Vector2 worldPosition)
    {
        var screenPosition = GetScreenPosition(worldPosition) + new Vector2(0f, comboBrokenTextOffsetY * Screen.height);
        AddGuiVanishingTextLocalized("IngameGui.ComboBroken", screenPosition, UnityHelper.AnchorCenter, negativeInfoColor, Font, fontSizeIngameInfo, infoTimeTillVanish, defaultVanishTime);
    }

    public void ShowComboOver(Vector2 worldPosition)
    {
        var screenPosition = GetScreenPosition(worldPosition);
        AddGuiVanishingTextLocalized("IngameGui.ComboOver", screenPosition, UnityHelper.AnchorCenter, positiveInfo, Font, fontSizeIngameInfo, infoTimeTillVanish, defaultVanishTime);
    }

    public void ShowHealthWarning(float healthLevel, Vector2 worldPosition)
    {
        var text = Get("IngameGui.HealthWarning", (int) (healthLevel * 100));

        var screenPosition = GetScreenPosition(worldPosition);
        AddGuiVanishingText(text, screenPosition, UnityHelper.AnchorCenter, negativeInfoColor, Font, fontSizeIngameInfo, infoTimeTillVanish, defaultVanishTime);
    }

    public void ShowDeathMessage(Vector2 worldPosition)
    {
        var screenPosition = GetScreenPosition(worldPosition);
        AddGuiVanishingTextLocalized("IngameGui.DeathMessage", screenPosition, UnityHelper.AnchorCenter, negativeInfoColor, Font, fontSizeIngameInfo, infoTimeTillVanish, defaultVanishTime);
    }

    private void AddGuiVanishingText(string text, Vector2 position, Vector2 relativeAnchor, Color color, Font font, float fontSize, float timeTillVanish, float vanishTime, TextAlignment alignment = TextAlignment.Left)
    {
        texts.Add(new GuiVanishingText(text, position, relativeAnchor, color, font, fontSize, timeTillVanish, vanishTime, alignment));
    }

    private void AddGuiVanishingTextLocalized(string localizedKey, Vector2 position, Vector2 relativeAnchor, Color color, Font font, float fontSize, float timeTillVanish, float vanishTime, TextAlignment alignment = TextAlignment.Left)
    {
        AddGuiVanishingText(Get(localizedKey), position, relativeAnchor, color, font, fontSize, timeTillVanish, vanishTime, alignment);
    }

    public Vector2 GetScreenPosition(Vector2 worldPosition)
    {
        var screenPosition = mainCamera.WorldToScreenPoint(worldPosition);
        screenPosition.y = Screen.height - screenPosition.y;
        return screenPosition;
    }

    public void Clear()
    {
        texts.Clear();
    }

    public void SetTotalScoreScreenPosition(Vector2 screenPosition)
    {
        totalScoreGUIPosition = screenPosition;
        totalScoreGUIPosition.y = Screen.height - screenPosition.y;
    }

    public void SetTotalScoreWorldPosition(Vector2 worldPosition)
    {
        SetTotalScoreScreenPosition(mainCamera.WorldToScreenPoint(worldPosition));
    }

    public Rect ShowTextLocalized(string localizationKey, Vector2 screenPosition, Vector2 relativeAnchor, Color color, TextAlignment alignment = TextAlignment.Left)
    {
        return UnityHelper.ShowText(localizationKey, screenPosition, relativeAnchor, color, alignment);
    }

    public string Get(string localizationKey, params object[] args)
    {
        return String.Format(localizationKey, args);
    }

    public void MouseOverButtonAreaChange(GuiButton guiButton, bool inArea)
    {
        if (inArea)
        {
            mouseOverButtons.Add(guiButton);
        }
        else
        {
            mouseOverButtons.Remove(guiButton);
        }

        if (mouseOverButtons.Count == 0)
        {
            CurrentlyHoveredButton = null;
        }
        else if (mouseOverButtons.Count == 1)
        {
            CurrentlyHoveredButton = mouseOverButtons[0];
        }
        else
        {
            UpdateHoveredMouseButtonAmongstMultiple();
        }
    }

    private void UpdateHoveredMouseButtonAmongstMultiple()
    {
        Vector2 mousePosition = Input.mousePosition;
        CurrentlyHoveredButton = mouseOverButtons.OrderBy(button => Vector2.Distance(button.ScreenRect.center, mousePosition)).FirstOrDefault();
    }

    private GuiButton CurrentlyHoveredButton
    {
        get { return _currentlyHoveredButton; }
        set
        {
            if (_currentlyHoveredButton == value)
                return;

            if (_currentlyHoveredButton != null)
                _currentlyHoveredButton.Hovering = false;

            _currentlyHoveredButton = value;

            if (_currentlyHoveredButton != null)
                _currentlyHoveredButton.Hovering = true;
        }
    }
}
